# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details:
# <https://github.com/moble/quaternionic/blob/master/LICENSE>

import abc
import numpy as np
from . import jit
from .utilities import type_self_return, ndarray_args, ndarray_args_and_return

def QuaternionPropertiesMixin(jit=jit):
    class mixin(abc.ABC):
        """Basic properties for quaternionic array class

        This abstract base class comprises the basic interface for quaternionic
        arrays, including the components (w, x, y, z), the parts (scalar,
        vector), the norm and absolute value, the normalized equivalent, and so
        on.  Also included are various

        The other main elements for a quaternionic array class are converters
        (to and from matrices, Euler angles, etc.)  and, of course, the
        algebraic functions (addition, multiplication, etc.).

        """

        @property
        def w(self):
            """The first (scalar) component of the quaternion"""
            return self.ndarray[..., 0]

        @w.setter
        def w(self, wprime):
            self.ndarray[..., 0] = wprime

        @property
        def x(self):
            """The second component of the quaternion (rotation about x)"""
            return self.ndarray[..., 1]

        @x.setter
        def x(self, xprime):
            self.ndarray[..., 1] = xprime

        @property
        def y(self):
            """The third component of the quaternion (rotation about y)"""
            return self.ndarray[..., 2]

        @y.setter
        def y(self, yprime):
            self.ndarray[..., 2] = yprime

        @property
        def z(self):
            """The fourth component of the quaternion (rotation about z)"""
            return self.ndarray[..., 3]

        @z.setter
        def z(self, zprime):
            self.ndarray[..., 3] = zprime

        @property
        def vector(self):
            """The "vector" part of the quaternion (final three components)

            Note that it is entirely standard to describe this part of the
            quaternion as the "vector" part.  It would be more correct to refer
            to it as the "bivector" part, as explained by geometric algebra.

            """
            return self.ndarray[..., 1:]

        @vector.setter
        def vector(self, v):
            self.ndarray[..., 1:] = v

        @property
        @ndarray_args
        @jit
        def norm(self):
            """The (Cayley) norm of the quaternion

            This quantity is the sum of the squares of the components of the
            quaternion — equal to the square of the absolute value.

            Note that it may be surprising to find that this "norm" does not
            include the usual square root.  This is conventional.  For example,
            the Boost library's implementation of quaternions also uses this
            convention.  Similarly, the implementation of complex numbers in
            C++ defines the `abs` and `norm` functions in this way, while
            python and numpy only define `abs`.

            If you are uncomfortable with this choice of the meaning of `norm`,
            it may make more sense to use one of the aliases of this function,
            which include

              * absolute_square
              * abs2
              * mag2

            """
            s = self.reshape((-1, 4))
            n = np.empty(s.shape[0], dtype=self.dtype)
            for i in range(s.shape[0]):
                n[i] = s[i, 0]**2 + s[i, 1]**2 + s[i, 2]**2 + s[i, 3]**2
            return n.reshape(self.shape[:-1])

        @property
        @ndarray_args
        @jit
        def abs(self):
            """The absolute value of the quaternion

            This quantity is the square-root of the sum of the squares of the
            components of each quaternion.

            See the note in `norm` for the difference between this function and
            that.  Basically, this is the square-root of that function.

            """
            s = self.reshape((-1, 4))
            n = np.empty(s.shape[0], dtype=self.dtype)
            for i in range(s.shape[0]):
                n[i] = np.sqrt(s[i, 0]**2 + s[i, 1]**2 + s[i, 2]**2 + s[i, 3]**2)
            return n.reshape(self.shape[:-1])

        def conjugate(self):
            """The quaternion conjugate of this quaternion"""
            c = self.copy()
            c.vector *= -1
            return c

        @property
        @ndarray_args_and_return
        @jit
        def inverse(self):
            """The multiplicative inverse of this quaternion"""
            s = self.reshape((-1, 4))
            inv = np.empty(s.shape, dtype=self.dtype)
            for i in range(s.shape[0]):
                n = s[i, 0]**2 + s[i, 1]**2 + s[i, 2]**2 + s[i, 3]**2
                inv[i, 0] = s[i, 0] / n
                inv[i, 1] = -s[i, 1] / n
                inv[i, 2] = -s[i, 2] / n
                inv[i, 3] = -s[i, 3] / n
            return inv.reshape(self.shape)

        # Aliases
        scalar = w
        real = w
        i = x
        j = y
        k = z
        imag = vector
        absolute_square = norm
        abs2 = norm
        mag2 = norm
        modulus = abs
        magnitude = abs
        conj = conjugate

        @property
        def normalized(self):
            """The normalized version of this quaternion"""
            return self / self.abs

        @property
        def ndarray(self):
            """View this array as a numpy ndarray"""
            return self.view(np.ndarray)

        @property
        def flattened(self):
            """A view of this array with all but the last dimension combined into one"""
            return self.reshape((-1, 4))

        @property
        def iterator(self):
            """Iterate over all but the last dimension of this quaternion array"""
            s = self.reshape((-1, 4))
            for i in range(s.shape[0]):
                yield(s[i])

        def nonzero(self):
            """Return the indices of all nonzero elements

            This is essentially the same function as numpy.nonzero, except that
            the last dimension is treated as a single quaternion; if any
            component of the quaternion is nonzero, the quaternion is
            considered nonzero.

            """
            return np.nonzero(np.atleast_1d(np.any(self.ndarray, axis=-1)))

        def rotate(self, v, axis=-1):
            """Rotate vectors by quaternions in this array

            For simplicity, this function simply converts the input
            quaternion(s) to a matrix, and rotates the input vector(s) by the
            usual matrix multiplication.  However, it should be noted that if
            each input quaternion is only used to rotate a single vector, it is
            more efficient (in terms of operation counts) to use the formula

              v' = v + 2 * r x (s * v + r x v) / m

            where x represents the cross product, s and r are the scalar and
            vector parts of the quaternion, respectively, and m is the sum of
            the squares of the components of the quaternion.  If you are
            looping over a very large number of quaternions, and just rotating
            a single vector each time, you might want to implement that
            alternative algorithm using numba (or something that doesn't use
            python).


            Parameters
            ==========
            v: float array
                Three-vectors to be rotated.
            axis: int
                Axis of the `v` array to use as the vector dimension.  This axis of
                `v` must have length 3.

            Returns
            =======
            vprime: float array
                The rotated vectors.  This array has shape self.shape+v.shape.

            """
            v = np.asarray(v, dtype=self.dtype)
            if v.ndim < 1 or 3 not in v.shape:
                raise ValueError("Input `v` does not have at least one dimension of length 3")
            if v.shape[axis] != 3:
                raise ValueError("Input `v` axis {0} has length {1}, not 3.".format(axis, v.shape[axis]))
            m = self.to_rotation_matrix
            m_axes = list(range(m.ndim))
            v_axes = list(range(m.ndim, m.ndim+v.ndim))
            mv_axes = list(v_axes)
            mv_axes[axis] = m_axes[-2]
            mv_axes = m_axes[:-2] + mv_axes
            v_axes[axis] = m_axes[-1]
            return np.einsum(m, m_axes, v, v_axes, mv_axes)

    return mixin



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details:
# <https://github.com/moble/quaternionic/blob/master/LICENSE>

"""Essential functions for quaternion algebra

These functions form the basic algebraic behavior of quaternions — addition,
multiplication, exp, log, etc.

Each function takes one or two array-like inputs — depending on whether it is
unary or binary — as the first parameter or two, and another array-like object
for output as the final parameter.  Even for functions that return a single
float or bool, the output must be array-like so that it can be modified inside
the function.  These functions are JIT-compiled by numba's `guvectorize`
function, meaning that they can also act on arbitrary arrays just like standard
numpy arrays, as long as the final dimension of any quaternion-valued input has
size 4 to represent the components of the quaternion.

These functions are generic, meaning that they can be used without the
`quaternionic.array` object.  However, these functions are implemented as the
"ufunc"s of those arrays, meaning that we can perform algebra directly on them
in natural ways — as in `q1+q2`, `q1*q2`, etc. — and using the standard numpy
functions — as in `np.exp(q)`, `np.log(q)`, etc.

For this purpose, we implement as many of the standard ufuncs
<https://numpy.org/doc/stable/reference/ufuncs.html> as make sense for
quaternions.  For the most part, this means ignoring operations related to
integers, remainders, ordering, or trigonometric functions.  The bit-twiddling
functions are re-interpreted as they usually are in geometric algebra to denote
geometric operations.

All functions in this module are magically compiled to ufuncs in `__init__.py`
and placed attached to the `algebra_ufuncs` object.

"""


import numpy as np
from . import float64, boolean
from .utilities import attach_typelist_and_signature


_quaternion_resolution = 10 * np.finfo(float).resolution


@attach_typelist_and_signature([(float64[:], float64[:], float64[:])], '(n),(n)->(n)')
def add(q1, q2, qout):
    """Add two quaternions q1+q2"""
    qout[0] = q1[0] + q2[0]
    qout[1] = q1[1] + q2[1]
    qout[2] = q1[2] + q2[2]
    qout[3] = q1[3] + q2[3]


@attach_typelist_and_signature([(float64[:], float64[:], float64[:])], '(n),(n)->(n)')
def subtract(q1, q2, qout):
    """Subtract quaternion q1-q2"""
    qout[0] = q1[0] - q2[0]
    qout[1] = q1[1] - q2[1]
    qout[2] = q1[2] - q2[2]
    qout[3] = q1[3] - q2[3]


@attach_typelist_and_signature([(float64[:], float64[:], float64[:])], '(n),(n)->(n)')
def multiply(q1, q2, qout):
    """Multiply quaternions q1*q2"""
    qout[0] = q1[0]*q2[0] - q1[1]*q2[1] - q1[2]*q2[2] - q1[3]*q2[3]
    qout[1] = q1[0]*q2[1] + q1[1]*q2[0] + q1[2]*q2[3] - q1[3]*q2[2]
    qout[2] = q1[0]*q2[2] - q1[1]*q2[3] + q1[2]*q2[0] + q1[3]*q2[1]
    qout[3] = q1[0]*q2[3] + q1[1]*q2[2] - q1[2]*q2[1] + q1[3]*q2[0]


@attach_typelist_and_signature([(float64[:], float64[:], float64[:])], '(n),(n)->(n)')
def divide(q1, q2, qout):
    """Divide quaternions q1/q2 = q1 * q2.inverse"""
    q2norm = q2[0]**2 + q2[1]**2 + q2[2]**2 + q2[3]**2
    qout[0] = (+q1[0]*q2[0] + q1[1]*q2[1] + q1[2]*q2[2] + q1[3]*q2[3]) / q2norm
    qout[1] = (-q1[0]*q2[1] + q1[1]*q2[0] - q1[2]*q2[3] + q1[3]*q2[2]) / q2norm
    qout[2] = (-q1[0]*q2[2] + q1[1]*q2[3] + q1[2]*q2[0] - q1[3]*q2[1]) / q2norm
    qout[3] = (-q1[0]*q2[3] - q1[1]*q2[2] + q1[2]*q2[1] + q1[3]*q2[0]) / q2norm


true_divide = divide


@attach_typelist_and_signature([(float64, float64[:], float64[:])], '(),(n)->(n)')
def multiply_scalar(s, q, qout):
    """Multiply scalar by quaternion s*q"""
    qout[0] = s * q[0]
    qout[1] = s * q[1]
    qout[2] = s * q[2]
    qout[3] = s * q[3]


@attach_typelist_and_signature([(float64, float64[:], float64[:])], '(),(n)->(n)')
def divide_scalar(s, q, qout):
    """Divide scalar by quaternion s/q = s * q.inverse"""
    qnorm = q[0]**2 + q[1]**2 + q[2]**2 + q[3]**2
    qout[0] = s * q[0] / qnorm
    qout[1] = -s * q[1] / qnorm
    qout[2] = -s * q[2] / qnorm
    qout[3] = -s * q[3] / qnorm


true_divide_scalar = divide_scalar


@attach_typelist_and_signature([(float64[:], float64, float64[:])], '(n),()->(n)')
def scalar_multiply(q, s, qout):
    """Multiply quaternion by scalar q*s"""
    qout[0] = q[0] * s
    qout[1] = q[1] * s
    qout[2] = q[2] * s
    qout[3] = q[3] * s


@attach_typelist_and_signature([(float64[:], float64, float64[:])], '(n),()->(n)')
def scalar_divide(q, s, qout):
    """Divide quaternion by scalar q/s"""
    qout[0] = q[0] / s
    qout[1] = q[1] / s
    qout[2] = q[2] / s
    qout[3] = q[3] / s


scalar_true_divide = scalar_divide


@attach_typelist_and_signature([(float64[:], float64[:])], '(n)->(n)')
def negative(q, qout):
    """Return negative quaternion -q"""
    qout[0] = -q[0]
    qout[1] = -q[1]
    qout[2] = -q[2]
    qout[3] = -q[3]


@attach_typelist_and_signature([(float64[:], float64[:])], '(n)->(n)')
def positive(q, qout):
    """Return input quaternion q"""
    qout[0] = q[0]
    qout[1] = q[1]
    qout[2] = q[2]
    qout[3] = q[3]


@attach_typelist_and_signature([(float64[:], float64, float64[:])], '(n),()->(n)')
def float_power(q, s, qout):
    """Raise quaternion to scalar power exp(log(q)*s)"""
    b = np.sqrt(q[1]**2 + q[2]**2 + q[3]**2)
    if np.abs(b) <= _quaternion_resolution * np.abs(q[0]):
        if q[0] < 0.0:
            if np.abs(q[0] + 1) > _quaternion_resolution:
                qout[0] = np.log(-q[0])
                qout[1] = np.pi
                qout[2] = 0.0
                qout[3] = 0.0
            else:
                qout[0] = 0.0
                qout[1] = np.pi
                qout[2] = 0.0
                qout[3] = 0.0
        else:
            qout[0] = np.log(q[0])
            qout[1] = 0.0
            qout[2] = 0.0
            qout[3] = 0.0
    else:
        v = np.arctan2(b, q[0])
        f = v / b
        qout[0] = np.log(q[0] * q[0] + b * b) / 2.0
        qout[1] = f * q[1]
        qout[2] = f * q[2]
        qout[3] = f * q[3]
    qout *= s
    vnorm = np.sqrt(qout[1]**2 + qout[2]**2 + qout[3]**2)
    if vnorm > _quaternion_resolution:
        e = np.exp(qout[0])
        qout[0] = e * np.cos(vnorm)
        qout[1:] *= e * np.sin(vnorm) / vnorm
    else:
        qout[0] = np.exp(qout[0])
        qout[1] = 0.0
        qout[2] = 0.0
        qout[3] = 0.0


@attach_typelist_and_signature([(float64[:], float64[:])], '(n)->()')
def absolute(q, qout):
    """Return absolute value of quaternion |q|"""
    qout[0] = np.sqrt(q[0]**2 + q[1]**2 + q[2]**2 + q[3]**2)


@attach_typelist_and_signature([(float64[:], float64[:])], '(n)->(n)')
def conj(q, qout):
    """Return quaternion-conjugate of quaternion q̄"""
    qout[0] = +q[0]
    qout[1] = -q[1]
    qout[2] = -q[2]
    qout[3] = -q[3]


conjugate = conj


@attach_typelist_and_signature([(float64[:], float64[:])], '(n)->(n)')
def exp(q, qout):
    """Return exponential of input quaternion exp(q)"""
    vnorm = np.sqrt(q[1]**2 + q[2]**2 + q[3]**2)
    if vnorm > _quaternion_resolution:
        s = np.sin(vnorm) / vnorm
        e = np.exp(q[0])
        qout[0] = e * np.cos(vnorm)
        qout[1] = e * s * q[1]
        qout[2] = e * s * q[2]
        qout[3] = e * s * q[3]
    else:
        qout[0] = np.exp(q[0])
        qout[1] = 0.0
        qout[2] = 0.0
        qout[3] = 0.0


@attach_typelist_and_signature([(float64[:], float64[:])], '(n)->(n)')
def log(q, qout):
    """Return logarithm of input quaternion log(q)"""
    b = np.sqrt(q[1]**2 + q[2]**2 + q[3]**2)
    if b <= _quaternion_resolution * np.abs(q[0]):
        if q[0] < 0.0:
            if np.abs(q[0] + 1) > _quaternion_resolution:
                qout[0] = np.log(-q[0])
                qout[1] = np.pi
                qout[2] = 0.0
                qout[3] = 0.0
            else:
                qout[0] = 0.0
                qout[1] = np.pi
                qout[2] = 0.0
                qout[3] = 0.0
        else:
            qout[0] = np.log(q[0])
            qout[1] = 0.0
            qout[2] = 0.0
            qout[3] = 0.0
    else:
        v = np.arctan2(b, q[0])
        f = v / b
        qout[0] = np.log(q[0] * q[0] + b * b) / 2.0
        qout[1] = f * q[1]
        qout[2] = f * q[2]
        qout[3] = f * q[3]


@attach_typelist_and_signature([(float64[:], float64[:])], '(n)->(n)')
def sqrt(q, qout):
    """Return square-root of input quaternion √q

    The general formula is

        √q = (|q| + q) / √(2|q| + 2q.w)

    """
    # √Q = (a + Q) / √(2*a + 2*Q[0])
    a = np.sqrt(q[0]**2 + q[1]**2 + q[2]**2 + q[3]**2)
    if np.abs(a + q[0]) < _quaternion_resolution * a:
        qout[0] = 0.0
        qout[1] = np.sqrt(a)
        qout[2] = 0.0
        qout[3] = 0.0
    else:
        c = np.sqrt(0.5 / (a + q[0]))
        qout[0] = (a + q[0]) * c
        qout[1] = q[1] * c
        qout[2] = q[2] * c
        qout[3] = q[3] * c


@attach_typelist_and_signature([(float64[:], float64[:])], '(n)->(n)')
def square(q, qout):
    """Return square of quaternion q*q"""
    qout[0] = q[0]**2 - q[1]**2 - q[2]**2 - q[3]**2
    qout[1] = 2*q[0]*q[1]
    qout[2] = 2*q[0]*q[2]
    qout[3] = 2*q[0]*q[3]


@attach_typelist_and_signature([(float64[:], float64[:])], '(n)->(n)')
def reciprocal(q, qout):
    """Return reciprocal (inverse) of quaternion q.inverse"""
    norm = q[0]**2 + q[1]**2 + q[2]**2 + q[3]**2
    qout[0] = q[0] / norm
    qout[1] = -q[1] / norm
    qout[2] = -q[2] / norm
    qout[3] = -q[3] / norm


@attach_typelist_and_signature([(float64[:], float64[:], float64[:])], '(n),(n)->(n)')
def bitwise_or(q1, q2, qout):
    """Return scalar product of quaternions q1|q2

    If we denote by <a>ₛ the grade-s component of the general multivector `a`,
    we can express this product as

        a | b = Σₛ,ᵣ <<a>ₛ <b>ᵣ>₀

    Note that this is different from the "Hestenes dot" product where the sum
    runs over s≠0 and r≠0; that is the product returned by `galgebra` using
    this operator.

    """
    qout[0] = q1[0]*q2[0] - q1[1]*q2[1] - q1[2]*q2[2] - q1[3]*q2[3]
    qout[1] = 0.0
    qout[2] = 0.0
    qout[3] = 0.0


@attach_typelist_and_signature([(float64[:], float64[:], float64[:])], '(n),(n)->(n)')
def bitwise_xor(q1, q2, qout):
    """Return outer product of quaternions q1^q2

    This is the generalized outer product of geometric algebra.  If we denote
    by <a>ₛ the grade-s component of the general multivector `a`, we can
    express this product as

        a ^ b = Σₛ,ᵣ <<a>ₛ <b>ᵣ>ₛ₊ᵣ

    Note that the result may seem surprising because we sometimes think of quaternions as 

    """
    qout[0] = q1[0]*q2[0]
    qout[1] = q1[0]*q2[1] + q1[1]*q2[0]
    qout[2] = q1[0]*q2[2] + q1[2]*q2[0]
    qout[3] = q1[0]*q2[3] + q1[3]*q2[0]


invert = conj  # reversion (= conjugate for quaternion algebra)


@attach_typelist_and_signature([(float64[:], float64[:], float64[:])], '(n),(n)->(n)')
def left_shift(q1, q2, qout):
    """Return left-contraction of quaternions q1<<q2 = q1⌋q1

    For all quaternions a, b, c, we have

        (a ^ b) * c = a * (b ⌋ c)

    """
    qout[0] = q1[0]*q2[0] - q1[1]*q2[1] - q1[2]*q2[2] - q1[3]*q2[3]
    qout[1] = q1[0]*q2[1]
    qout[2] = q1[0]*q2[2]
    qout[3] = q1[0]*q2[3]


@attach_typelist_and_signature([(float64[:], float64[:], float64[:])], '(n),(n)->(n)')
def right_shift(q1, q2, qout):
    """Return right-contraction of quaternions q1>>q2 = q1⌊q2

    For all quaternions a, b, c, we have

        c * (b ^ a) = (c ⌊ b) * a

    """
    qout[0] = q1[0]*q2[0] - q1[1]*q2[1] - q1[2]*q2[2] - q1[3]*q2[3]
    qout[1] = q1[1]*q2[0]
    qout[2] = q1[2]*q2[0]
    qout[3] = q1[3]*q2[0]


@attach_typelist_and_signature([(float64[:], float64[:], boolean[:])], '(n),(n)->()')
def not_equal(q1, q2, bout):
    bout[0] = np.any(q1[:] != q2[:])


@attach_typelist_and_signature([(float64[:], float64[:], boolean[:])], '(n),(n)->()')
def equal(q1, q2, bout):
    bout[0] = np.all(q1[:] == q2[:])


@attach_typelist_and_signature([(float64[:], float64[:], boolean[:])], '(n),(n)->()')
def logical_and(q1, q2, bout):
    bout[0] = np.any(q1[:]) and np.any(q2[:])


@attach_typelist_and_signature([(float64[:], float64[:], boolean[:])], '(n),(n)->()')
def logical_or(q1, q2, bout):
    bout[0] = np.any(q1[:]) or np.any(q2[:])


@attach_typelist_and_signature([(float64[:], boolean[:])], '(n)->()')
def isfinite(qin, bout):
    bout[0] = np.isfinite(qin[0]) and np.isfinite(qin[1]) and np.isfinite(qin[2]) and np.isfinite(qin[3])


@attach_typelist_and_signature([(float64[:], boolean[:])], '(n)->()')
def isinf(qin, bout):
    bout[0] = np.isinf(qin[0]) or np.isinf(qin[1]) or np.isinf(qin[2]) or np.isinf(qin[3])


@attach_typelist_and_signature([(float64[:], boolean[:])], '(n)->()')
def isnan(qin, bout):
    bout[0] = np.isnan(qin[0]) or np.isnan(qin[1]) or np.isnan(qin[2]) or np.isnan(qin[3])



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details:
# <https://github.com/moble/quaternionic/blob/master/LICENSE>

import sys
import functools

ufunc_attributes = [
    'nin', 'nout', 'nargs', 'ntypes', 'types', 'identity', 'signature',
    'reduce', 'accumulate', 'reduceat', 'outer', 'at'
]


def type_self_return(f):
    """Decorate jitted functions to return with type of first argument"""
    @functools.wraps(f)
    def f_wrapped(*args, **kwargs):
        return_value = f(*args, **kwargs)
        return type(args[0])(return_value)
    for attr in ufunc_attributes:
        if hasattr(f, attr):
            setattr(f_wrapped, attr, getattr(f, attr))
    return f_wrapped


def ndarray_args(f):
    """Decorate jitted functions to accept quaternionic arrays"""
    @functools.wraps(f)
    def f_wrapped(*args, **kwargs):
        args_ndarray = (arg if not hasattr(arg, 'ndarray') else arg.ndarray for arg in args)
        return f(*args_ndarray, **kwargs)
    for attr in ufunc_attributes:
        if hasattr(f, attr):
            setattr(f_wrapped, attr, getattr(f, attr))
    return f_wrapped


def ndarray_args_and_return(f):
    """Decorate jitted functions to accept and return quaternionic arrays"""
    @functools.wraps(f)
    def f_wrapped(*args, **kwargs):
        args_ndarray = (arg if not hasattr(arg, 'ndarray') else arg.ndarray for arg in args)
        return type(args[0])(f(*args_ndarray, **kwargs))
    for attr in ufunc_attributes:
        if hasattr(f, attr):
            setattr(f_wrapped, attr, getattr(f, attr))
    return f_wrapped


def convert_numpy_ufunc_type_to_numba_ftylist(typelist):
    import numba
    # https://numpy.org/doc/stable/reference/arrays.scalars.html#arrays-scalars-character-codes
    # https://numba.pydata.org/numba-doc/latest/reference/types.html
    map_numpy_typecode_to_numba_type = {
        '?': numba.boolean,
        'b': numba.byte,
        'h': numba.short,
        'i': numba.intc,
        'l': numba.int_,
        'q': numba.longlong,
        'p': numba.intp,
        # 'B': numba.ubyte,  # numba.ubyte doesn't exist
        'B': numba.char,  # probably closest
        'H': numba.ushort,
        'I': numba.uintc,
        'L': numba.uint,
        'Q': numba.ulonglong,
        'P': numba.uintp,
        # 'e': numba.half,  # numba.half doesn't exist
        # 'f': numba.single,  # numba.single doesn't exist
        'f': numba.float32,  # probably closest
        'd': numba.float_,
        # 'g': numba.longfloat,  # numba.longfloat doesn't exist
        'g': numba.double,  # probably closest
        # 'F': numba.csingle,  # numba.csingle doesn't exist
        'F': numba.complex64,  # probably closest
        # 'D': numba.complex_,  # numba.complex_ doesn't exist
        'D': numba.complex128,  # probably closest
        # 'G': numba.clongfloat,  # numba.clongfloat doesn't exist
    }

    ftylist = []
    for types in typelist:
        in_types, out_type = types.split('->')
        inputs = (map_numpy_typecode_to_numba_type[c] for c in in_types)
        output = map_numpy_typecode_to_numba_type[out_type]
        ftylist.append(output(*inputs))
    return ftylist


def attach_typelist_and_signature(ftylist, signature):
    import functools
    def wrapper(f):
        f.types = ftylist
        f.signature = signature
        return f
    return wrapper


def guvectorize_module_functions(module, obj):
    """Compile all functions in module to ufuncs and attach to obj

    Note that the functions in module must have attributes `types` and
    `signature`, providing the necessary arguments to numba's guvectorize
    decorator; any function that does not have those attributes will be skipped
    silently.

    """
    import types
    for k, v in module.__dict__.items():
        if isinstance(v, types.FunctionType) and hasattr(v, 'types') and hasattr(v, 'signature'):
            v_ufunc = guvectorize(v.types, v.signature)(v)
            setattr(obj, k, v_ufunc)


def pyguvectorize(types, signature):
    """Test function to convert functions to general universal functions

    Note that this is almost certainly only useful for functions defined in
    quaternionic.algebra — and specifically only if they have `type` and
    `signature` attributes.  Moreover, the result is a slow python function,
    meant only for testing.


    Nonetheless, the output of this decorator should be callable with the same
    signature as a gufunc.  In this way, tests designed for the compiled
    gufuncs can also be used to test the python-mode code and obtain coverage
    results.

    """
    import functools
    import numpy as np
    inputs, output = signature.split('->')
    inputs = inputs.split(',')
    slice_a = slice(None) if inputs[0]=='()' else 0
    pad_a = np.newaxis if inputs[0]=='()' else slice(None)
    last_dim_a = slice(None, 1) if inputs[0]=='()' else slice(None)
    if len(inputs) > 1:
        slice_b = slice(None) if inputs[1]=='()' else 0
        pad_b = np.newaxis if inputs[1]=='()' else slice(None)
        last_dim_b = slice(None, 1) if inputs[1]=='()' else  slice(None)
    #slice_c = slice(None) if output=='()' else 0
    pad_c = slice(None)# np.newaxis if output=='()' else slice(None)
    last_dim_c = slice(None, 1) if output=='()' else  slice(None)
    last_axis_c = tuple() if output=='()' else (4,)
    dtype_c = np.dtype(types[0][-1].dtype.name)
    if len(inputs) == 1:
        def wrapper(f):
            @functools.wraps(f)
            def f_wrapped(a):
                shape_c = a[..., slice_a].shape + last_axis_c
                c = np.empty(shape_c, dtype=dtype_c)
                if not last_axis_c:
                    c = c[..., np.newaxis]
                a, ctmp = np.broadcast_arrays(a[..., pad_a], c[..., pad_c])
                a = a.reshape((-1, 4))
                ctmp.flags.writeable = True
                ctmp = ctmp.reshape((-1, 4))
                for a_i, c_i in zip(a, ctmp):
                    f(a_i, c_i)
                return c.reshape(shape_c)
            return f_wrapped
    else:
        def wrapper(f):
            @functools.wraps(f)
            def f_wrapped(a, b):
                shape_c = np.broadcast(a[..., slice_a], b[..., slice_b]).shape + last_axis_c
                c = np.empty(shape_c, dtype=dtype_c)
                if not last_axis_c:
                    c = c[..., np.newaxis]
                a, b, ctmp = np.broadcast_arrays(a[..., pad_a], b[..., pad_b], c[..., pad_c])
                a = a.reshape((-1, 4))
                b = b.reshape((-1, 4))
                ctmp = ctmp.reshape((-1, 4))
                ctmp.flags.writeable = True
                for a_i, b_i, c_i in zip(a, b, ctmp):
                    f(a_i[last_dim_a], b_i[last_dim_b], c_i[last_dim_c])
                return c.reshape(shape_c)
            return f_wrapped
    return wrapper


def pyguvectorize_module_functions(module, obj):
    """Wrap all functions in module to ufunc-like python funcs and attach to obj

    This function is almost exclusively useful for testing purposes.  See
    docstring of pyguvectorize function for details.

    """
    import types
    for k, v in module.__dict__.items():
        if isinstance(v, types.FunctionType) and hasattr(v, 'types') and hasattr(v, 'signature'):
            v_ufunc = pyguvectorize(v.types, v.signature)(v)
            setattr(obj, k, v_ufunc)


if sys.implementation.name.lower() == 'pypy':  # pragma: no cover
    class _FakeNumbaType(object):
        def __init__(self, name):
            self.name = name
        @property
        def dtype(self):
            return self
        def __getitem__(self, *args, **kwargs):
            return self
    float64 = _FakeNumbaType('float64')
    boolean = _FakeNumbaType('boolean')
    jit = lambda f: f
    guvectorize = pyguvectorize
else:
    import numba
    from numba import float64, boolean
    jit = functools.partial(numba.njit, cache=True)
    guvectorize = functools.partial(numba.guvectorize, nopython=True, cache=True)



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details:
# <https://github.com/moble/quaternionic/blob/master/LICENSE>

import numpy as np
from . import jit, guvectorize, algebra, float64
from .utilities import ndarray_args

_divide = jit(algebra.divide)
_log = jit(algebra.log)
_absolute = jit(algebra.absolute)


def CreateMetrics(jit=jit, guvectorize=guvectorize):
    class Rotor(object):
        @ndarray_args
        @guvectorize([(float64[:], float64[:], float64[:])], '(n),(n)->()')
        def intrinsic(q1, q2, out):
            """Geodesic distance between rotors within the Spin(3)=SU(2) manifold

            This function is equivalent to

                np.absolute(np.log(q1 / q2))

            which is a measure of the "distance" between two quaternions.  Note
            that no normalization is performed, which means that if q1 and/or
            q2 do not have unit norm, this is a more general type of distance.
            If they are normalized, the result of this function is half the
            angle through which vectors rotated by q1 would need to be rotated
            to lie on the same vectors rotated by q2.

            Parameters
            ----------
            q1, q2: QuaternionicArray
                Quaternionic arrays to be measured

            See also
            --------
            quaternionic.distance.rotor.chordal
            quaternionic.distance.rotation.intrinsic
            quaternionic.distance.rotation.chordal

            """
            qtemp = np.empty(4)
            _divide(q1, q2, qtemp)
            _log(qtemp, qtemp)
            _absolute(qtemp, out)
            out[0] *= 2

        @ndarray_args
        @guvectorize([(float64[:], float64[:], float64[:])], '(n),(n)->()')
        def chordal(q1, q2, out):
            """Euclidean distance between rotors

            This function is equivalent to

                np.absolute(q1 - q2)

            Note that no normalization is performed.  If the quaternions are
            normalized, this represents the length of the chord joining these
            two points on the unit 3-sphere, considered as embedded in
            Euclidean 4-space.

            Parameters
            ----------
            q1, q2: QuaternionicArray
                Quaternionic arrays to be measured

            See also
            --------
            quaternionic.distance.rotor.intrinsic
            quaternionic.distance.rotation.intrinsic
            quaternionic.distance.rotation.chordal

            """
            out[0] = np.sqrt((q1[0]-q2[0])**2 + (q1[1]-q2[1])**2 + (q1[2]-q2[2])**2 + (q1[3]-q2[3])**2)


    class Rotation(object):
        @ndarray_args
        @guvectorize([(float64[:], float64[:], float64[:])], '(n),(n)->()')
        def intrinsic(q1, q2, out):
            """Geodesic distance between rotations within the SO(3) manifold

            This function is equivalent to

                min(
                    np.absolute(np.log(q1 / q2)),
                    np.absolute(np.log(q1 / -q2))
                )

            which is a measure of the "distance" between two rotations.  Note
            that no normalization is performed, which means that if q1 and/or
            q2 do not have unit norm, this is a more general type of distance.
            If they are normalized, the result of this function is half the
            angle through which vectors rotated by q1 would need to be rotated
            to lie on the same vectors rotated by q2.

            Parameters
            ----------
            q1, q2: QuaternionicArray
                Quaternionic arrays to be measured

            See also
            --------
            quaternionic.distance.rotor.chordal
            quaternionic.distance.rotor.intrinsic
            quaternionic.distance.rotation.chordal

            """
            qtemp = np.empty(4)
            a = (q1[0]-q2[0])**2 + (q1[1]-q2[1])**2 + (q1[2]-q2[2])**2 + (q1[3]-q2[3])**2
            b = (q1[0]+q2[0])**2 + (q1[1]+q2[1])**2 + (q1[2]+q2[2])**2 + (q1[3]+q2[3])**2
            if b > a:
                _divide(q1, q2, qtemp)
            else:
                _divide(q1, -q2, qtemp)
            _log(qtemp, qtemp)
            _absolute(qtemp, out)
            out[0] *= 2

        @ndarray_args
        @guvectorize([(float64[:], float64[:], float64[:])], '(n),(n)->()')
        def chordal(q1, q2, out):
            """Euclidean distance between rotations

            This function is equivalent to

                min(
                    np.absolute(q1 - q2),
                    np.absolute(q1 + q2)
                )

            Note that no normalization is performed.  If the quaternions are
            normalized, this represents the length of the chord joining these
            two points on the unit 3-sphere, considered as embedded in
            Euclidean 4-space.

            Parameters
            ----------
            q1, q2: QuaternionicArray
                Quaternionic arrays to be measured

            See also
            --------
            quaternionic.distance.rotor.intrinsic
            quaternionic.distance.rotor.chordal
            quaternionic.distance.rotation.intrinsic

            """
            a = (q1[0]-q2[0])**2 + (q1[1]-q2[1])**2 + (q1[2]-q2[2])**2 + (q1[3]-q2[3])**2
            b = (q1[0]+q2[0])**2 + (q1[1]+q2[1])**2 + (q1[2]+q2[2])**2 + (q1[3]+q2[3])**2
            if b > a:
                out[0] = np.sqrt(a)
            else:
                out[0] = np.sqrt(b)


    return Rotor, Rotation


rotor, rotation = CreateMetrics()



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details:
# <https://github.com/moble/quaternionic/blob/master/LICENSE>

import abc
import numpy as np
from . import jit
from .utilities import ndarray_args


def QuaternionConvertersMixin(jit=jit):
    class mixin(abc.ABC):
        """Converters for quaternionic array class

        This abstract base class provides converters for quaternionic arrays, going
        to and from other representations, including the axis-angle representation,
        rotation matrices, Euler angles, spherical coordinates.

        """

        @property
        @ndarray_args
        @jit
        def to_rotation_matrix(self):
            """Convert quaternions to 3x3 rotation matrices

            Assuming the quaternion R rotates a vector v according to

                v' = R * v * R⁻¹,

            we can also express this rotation in terms of a 3x3 matrix ℛ such that

                v' = ℛ * v.

            This function returns that matrix.


            Returns
            -------
            rot: float array
                Output shape is self.shape[:-1]+(3,3).  This matrix should multiply
                (from the left) a column vector to produce the rotated column
                vector.

            """
            s = self.reshape((-1, 4))
            m = np.empty(s.shape[:1] + (3, 3), dtype=self.dtype)
            for i in range(s.shape[0]):
                n = s[i, 0]**2 + s[i, 1]**2 + s[i, 2]**2 + s[i, 3]**2
                m[i, 0, 0] = 1.0 - 2*(s[i, 2]**2 + s[i, 3]**2) / n
                m[i, 0, 1] = 2*(s[i, 1]*s[i, 2] - s[i, 3]*s[i, 0]) / n
                m[i, 0, 2] = 2*(s[i, 1]*s[i, 3] + s[i, 2]*s[i, 0]) / n
                m[i, 1, 0] = 2*(s[i, 1]*s[i, 2] + s[i, 3]*s[i, 0]) / n
                m[i, 1, 1] = 1.0 - 2*(s[i, 1]**2 + s[i, 3]**2) / n
                m[i, 1, 2] = 2*(s[i, 2]*s[i, 3] - s[i, 1]*s[i, 0]) / n
                m[i, 2, 0] = 2*(s[i, 1]*s[i, 3] - s[i, 2]*s[i, 0]) / n
                m[i, 2, 1] = 2*(s[i, 2]*s[i, 3] + s[i, 1]*s[i, 0]) / n
                m[i, 2, 2] = 1.0 - 2*(s[i, 1]**2 + s[i, 2]**2) / n
            return m.reshape(self.shape[:-1] + (3, 3))

        @classmethod
        def from_rotation_matrix(cls, rot, nonorthogonal=True):
            """Convert input 3x3 rotation matrix to unit quaternion

            By default, if scipy.linalg is available, this function uses
            Bar-Itzhack's algorithm to allow for non-orthogonal matrices.
            [J. Guidance, Vol. 23, No. 6, p. 1085 <http://dx.doi.org/10.2514/2.4654>]
            This will almost certainly be quite a bit slower than simpler versions,
            though it will be more robust to numerical errors in the rotation matrix.
            Also note that Bar-Itzhack uses some pretty weird conventions.  The last
            component of the quaternion appears to represent the scalar, and the
            quaternion itself is conjugated relative to the convention used
            throughout this module.

            If scipy.linalg is not available or if the optional
            `nonorthogonal` parameter is set to `False`, this function falls
            back to the possibly faster, but less robust, algorithm of Markley
            [J. Guidance, Vol. 31, No. 2, p. 440
            <http://dx.doi.org/10.2514/1.31730>].

            Parameters
            ----------
            rot: (...Nx3x3) float array
                Each 3x3 matrix represents a rotation by multiplying (from the left)
                a column vector to produce a rotated column vector.  Note that this
                input may actually have ndims>3; it is just assumed that the last
                two dimensions have size 3, representing the matrix.
            nonorthogonal: bool, optional
                If scipy.linalg is available, use the more robust algorithm of
                Bar-Itzhack.  Default value is True.

            Returns
            -------
            q: array of quaternions
                Unit quaternions resulting in rotations corresponding to input
                rotations.  Output shape is rot.shape[:-2].

            Raises
            ------
            LinAlgError
                If any of the eigenvalue solutions does not converge

            """
            from scipy import linalg

            rot = np.array(rot, copy=False)
            shape = rot.shape[:-2]

            if nonorthogonal:
                from operator import mul
                from functools import reduce

                K3 = np.empty(shape+(4, 4), dtype=rot.dtype)
                K3[..., 0, 0] = (rot[..., 0, 0] - rot[..., 1, 1] - rot[..., 2, 2])/3
                K3[..., 0, 1] = (rot[..., 1, 0] + rot[..., 0, 1])/3
                K3[..., 0, 2] = (rot[..., 2, 0] + rot[..., 0, 2])/3
                K3[..., 0, 3] = (rot[..., 1, 2] - rot[..., 2, 1])/3
                K3[..., 1, 0] = K3[..., 0, 1]
                K3[..., 1, 1] = (rot[..., 1, 1] - rot[..., 0, 0] - rot[..., 2, 2])/3
                K3[..., 1, 2] = (rot[..., 2, 1] + rot[..., 1, 2])/3
                K3[..., 1, 3] = (rot[..., 2, 0] - rot[..., 0, 2])/3
                K3[..., 2, 0] = K3[..., 0, 2]
                K3[..., 2, 1] = K3[..., 1, 2]
                K3[..., 2, 2] = (rot[..., 2, 2] - rot[..., 0, 0] - rot[..., 1, 1])/3
                K3[..., 2, 3] = (rot[..., 0, 1] - rot[..., 1, 0])/3
                K3[..., 3, 0] = K3[..., 0, 3]
                K3[..., 3, 1] = K3[..., 1, 3]
                K3[..., 3, 2] = K3[..., 2, 3]
                K3[..., 3, 3] = (rot[..., 0, 0] + rot[..., 1, 1] + rot[..., 2, 2])/3

                if not shape:
                    q = np.empty((4,), dtype=rot.dtype)
                    eigvals, eigvecs = linalg.eigh(K3.T, eigvals=(3, 3))
                    q[0] = eigvecs[-1]
                    q[1:] = -eigvecs[:-1].flatten()
                    return cls(q)
                else:
                    q = np.empty(shape+(4,), dtype=rot.dtype)
                    for flat_index in range(reduce(mul, shape)):
                        multi_index = np.unravel_index(flat_index, shape)
                        eigvals, eigvecs = linalg.eigh(K3[multi_index], eigvals=(3, 3))
                        q[multi_index+(0,)] = eigvecs[-1]
                        q[multi_index+(slice(1,None),)] = -eigvecs[:-1].flatten()
                    return cls(q)

            else:  # Not `nonorthogonal`
                diagonals = np.empty(shape+(4,), dtype=rot.dtype)
                diagonals[..., 0] = rot[..., 0, 0]
                diagonals[..., 1] = rot[..., 1, 1]
                diagonals[..., 2] = rot[..., 2, 2]
                diagonals[..., 3] = rot[..., 0, 0] + rot[..., 1, 1] + rot[..., 2, 2]

                indices = np.argmax(diagonals, axis=-1)

                q = diagonals  # reuse storage space
                indices_i = (indices == 0)
                if np.any(indices_i):
                    if indices_i.shape == ():
                        indices_i = Ellipsis
                    rot_i = rot[indices_i, :, :]
                    q[indices_i, 0] = rot_i[..., 2, 1] - rot_i[..., 1, 2]
                    q[indices_i, 1] = 1 + rot_i[..., 0, 0] - rot_i[..., 1, 1] - rot_i[..., 2, 2]
                    q[indices_i, 2] = rot_i[..., 0, 1] + rot_i[..., 1, 0]
                    q[indices_i, 3] = rot_i[..., 0, 2] + rot_i[..., 2, 0]
                indices_i = (indices == 1)
                if np.any(indices_i):
                    if indices_i.shape == ():
                        indices_i = Ellipsis
                    rot_i = rot[indices_i, :, :]
                    q[indices_i, 0] = rot_i[..., 0, 2] - rot_i[..., 2, 0]
                    q[indices_i, 1] = rot_i[..., 1, 0] + rot_i[..., 0, 1]
                    q[indices_i, 2] = 1 - rot_i[..., 0, 0] + rot_i[..., 1, 1] - rot_i[..., 2, 2]
                    q[indices_i, 3] = rot_i[..., 1, 2] + rot_i[..., 2, 1]
                indices_i = (indices == 2)
                if np.any(indices_i):
                    if indices_i.shape == ():
                        indices_i = Ellipsis
                    rot_i = rot[indices_i, :, :]
                    q[indices_i, 0] = rot_i[..., 1, 0] - rot_i[..., 0, 1]
                    q[indices_i, 1] = rot_i[..., 2, 0] + rot_i[..., 0, 2]
                    q[indices_i, 2] = rot_i[..., 2, 1] + rot_i[..., 1, 2]
                    q[indices_i, 3] = 1 - rot_i[..., 0, 0] - rot_i[..., 1, 1] + rot_i[..., 2, 2]
                indices_i = (indices == 3)
                if np.any(indices_i):
                    if indices_i.shape == ():
                        indices_i = Ellipsis
                    rot_i = rot[indices_i, :, :]
                    q[indices_i, 0] = 1 + rot_i[..., 0, 0] + rot_i[..., 1, 1] + rot_i[..., 2, 2]
                    q[indices_i, 1] = rot_i[..., 2, 1] - rot_i[..., 1, 2]
                    q[indices_i, 2] = rot_i[..., 0, 2] - rot_i[..., 2, 0]
                    q[indices_i, 3] = rot_i[..., 1, 0] - rot_i[..., 0, 1]

                q /= np.linalg.norm(q, axis=-1)[..., np.newaxis]

                return cls(q)

        @property
        @ndarray_args
        @jit
        def to_transformation_matrix(self):
            """Convert quaternions to 4x4 transformation matrices

            Assuming the quaternion Q transforms another quaternion P according to

                P' = Q * P * Q̄,

            we can also express this rotation in terms of a 4x4 matrix ℳ such that

                P' = ℳ * P,

            where P is viewed as a 4-vector in the last line.  This function
            returns that matrix.


            Returns
            -------
            m: float array
                Output shape is self.shape[:-1]+(4,4).  This matrix should multiply
                (from the left) a column vector to produce the transformed column
                vector.

            See also
            --------
            to_rotation_matrix: assumes Q is a unit quaternion

            """
            s = self.reshape((-1, 4))
            m = np.zeros(s.shape[:1] + (4, 4), dtype=self.dtype)
            for i in range(s.shape[0]):
                m[i, 0, 0] = s[i, 0]**2 + s[i, 1]**2 + s[i, 2]**2 + s[i, 3]**2
                m[i, 1, 1] = s[i, 0]**2 + s[i, 1]**2 - s[i, 2]**2 - s[i, 3]**2
                m[i, 1, 2] = 2 * (s[i, 1]*s[i, 2] - s[i, 3]*s[i, 0])
                m[i, 1, 3] = 2 * (s[i, 1]*s[i, 3] + s[i, 2]*s[i, 0])
                m[i, 2, 1] = 2 * (s[i, 1]*s[i, 2] + s[i, 3]*s[i, 0])
                m[i, 2, 2] = s[i, 0]**2 - s[i, 1]**2 + s[i, 2]**2 - s[i, 3]**2
                m[i, 2, 3] = 2 * (s[i, 2]*s[i, 3] - s[i, 1]*s[i, 0])
                m[i, 3, 1] = 2 * (s[i, 1]*s[i, 3] - s[i, 2]*s[i, 0])
                m[i, 3, 2] = 2 * (s[i, 2]*s[i, 3] + s[i, 1]*s[i, 0])
                m[i, 3, 3] = s[i, 0]**2 - s[i, 1]**2 - s[i, 2]**2 + s[i, 3]**2
            return m.reshape(self.shape[:-1] + (4, 4))

        @property
        def to_axis_angle(self):
            """Convert input quaternion to the axis-angle representation

            Note that if any of the input quaternions has norm zero, no error is
            raised, but NaNs will appear in the output.

            Returns
            -------
            rot: float array
                Output shape is q.shape[:-1]+(3,).  Each vector represents the axis of
                the rotation, with norm proportional to the angle of the rotation in
                radians.

            """
            return 2 * np.log(self.normalized).vector

        to_rotation_vector = to_axis_angle

        @classmethod
        def from_axis_angle(cls, vec):
            """Convert 3-vector in axis-angle representation to unit quaternion

            Parameters
            ----------
            vec: (...N, 3) float array
                Each vector represents the axis of the rotation, with norm
                proportional to the angle of the rotation in radians.

            Returns
            -------
            q: array of quaternions
                Unit quaternions resulting in rotations corresponding to input
                rotations.  Output shape is rot.shape[:-1].

            """
            vec = np.asarray(vec)
            quats = np.zeros(vec.shape[:-1] + (4,), dtype=vec.dtype)
            quats[..., 1:] = vec[...] / 2
            return np.exp(cls(quats))

        from_rotation_vector = from_axis_angle

        @property
        @ndarray_args
        @jit
        def to_euler_angles(self):
            """Open Pandora's Box

            If somebody is trying to make you use Euler angles, tell them no, and
            walk away, and go and tell your mum.

            You don't want to use Euler angles.  They are awful.  Stay away.  It's
            one thing to convert from Euler angles to quaternions; at least you're
            moving in the right direction.  But to go the other way?!  It's just not
            right.

            Assumes the Euler angles correspond to the quaternion R via

                R = exp(alpha*z/2) * exp(beta*y/2) * exp(gamma*z/2)

            The angles are naturally in radians.

            NOTE: Before opening an issue reporting something "wrong" with this
            function, be sure to read all of the following page, *especially* the
            very last section about opening issues or pull requests.
            <https://github.com/moble/quaternion/wiki/Euler-angles-are-horrible>

            Returns
            -------
            alpha_beta_gamma: float array
                Output shape is q.shape+(3,).  These represent the angles (alpha,
                beta, gamma) in radians, where the normalized input quaternion
                represents `exp(alpha*z/2) * exp(beta*y/2) * exp(gamma*z/2)`.

            Raises
            ------
            AllHell
                ...if you try to actually use Euler angles, when you could have
                been using quaternions like a sensible person.

            """
            s = self.reshape((-1, 4))
            alpha_beta_gamma = np.empty((s.shape[0], 3), dtype=self.dtype)
            for i in range(s.shape[0]):
                n = s[i, 0]**2 + s[i, 1]**2 + s[i, 2]**2 + s[i, 3]**2
                alpha_beta_gamma[i, 0] = np.arctan2(s[i, 3], s[i, 0]) + np.arctan2(-s[i, 1], s[i, 2])
                alpha_beta_gamma[i, 1] = 2*np.arccos(np.sqrt((s[i, 0]**2 + s[i, 3]**2) / n))
                alpha_beta_gamma[i, 2] = np.arctan2(s[i, 3], s[i, 0]) - np.arctan2(-s[i, 1], s[i, 2])
            return alpha_beta_gamma.reshape(self.shape[:-1] + (3,))

        @classmethod
        def from_euler_angles(cls, alpha_beta_gamma, beta=None, gamma=None):
            """Improve your life drastically

            Assumes the Euler angles correspond to the quaternion R via

                R = exp(alpha*z/2) * exp(beta*y/2) * exp(gamma*z/2)

            The angles naturally must be in radians for this to make any sense.

            NOTE: Before opening an issue reporting something "wrong" with this
            function, be sure to read all of the following page, *especially* the
            very last section about opening issues or pull requests.
            <https://github.com/moble/quaternion/wiki/Euler-angles-are-horrible>

            Parameters
            ----------
            alpha_beta_gamma: float or array of floats
                This argument may either contain an array with last dimension of
                size 3, where those three elements describe the (alpha, beta, gamma)
                radian values for each rotation; or it may contain just the alpha
                values, in which case the next two arguments must also be given.
            beta: None, float, or array of floats
                If this array is given, it must be able to broadcast against the
                first and third arguments.
            gamma: None, float, or array of floats
                If this array is given, it must be able to broadcast against the
                first and second arguments.

            Returns
            -------
            R: quaternion array
                The shape of this array will be the same as the input, except that
                the last dimension will be removed.

            """
            # Figure out the input angles from either type of input
            if gamma is None:
                alpha_beta_gamma = np.asarray(alpha_beta_gamma)
                alpha = alpha_beta_gamma[..., 0]
                beta  = alpha_beta_gamma[..., 1]
                gamma = alpha_beta_gamma[..., 2]
            else:
                alpha = np.asarray(alpha_beta_gamma)
                beta  = np.asarray(beta)
                gamma = np.asarray(gamma)

            # Set up the output array
            R = np.empty(np.broadcast(alpha, beta, gamma).shape + (4,), dtype=alpha.dtype)

            # Compute the actual values of the quaternion components
            R[..., 0] =  np.cos(beta/2)*np.cos((alpha+gamma)/2)  # scalar quaternion components
            R[..., 1] = -np.sin(beta/2)*np.sin((alpha-gamma)/2)  # x quaternion components
            R[..., 2] =  np.sin(beta/2)*np.cos((alpha-gamma)/2)  # y quaternion components
            R[..., 3] =  np.cos(beta/2)*np.sin((alpha+gamma)/2)  # z quaternion components

            return cls(R)

        @property
        def to_spherical_coordinates(self):
            """Return the spherical coordinates corresponding to this quaternion

            Obviously, spherical coordinates do not contain as much information as a
            quaternion, so this function does lose some information.  However, the
            returned spherical coordinates will represent the point(s) on the sphere
            to which the input quaternion(s) rotate the z axis.

            Returns
            -------
            vartheta_varphi: float array
                Output shape is q.shape+(2,).  These represent the angles (vartheta,
                varphi) in radians, where the normalized input quaternion represents
                `exp(varphi*z/2) * exp(vartheta*y/2)`, up to an arbitrary inital
                rotation about `z`.

            """
            return self.to_euler_angles[..., 1::-1]

        @classmethod
        def from_spherical_coordinates(cls, theta_phi, phi=None):
            """Return the quaternion corresponding to these spherical coordinates

            Assumes the spherical coordinates correspond to the quaternion R via

                R = exp(phi*z/2) * exp(theta*y/2)

            The angles naturally must be in radians for this to make any sense.

            Note that this quaternion rotates `z` onto the point with the given
            spherical coordinates, but also rotates `x` and `y` onto the usual basis
            vectors (theta and phi, respectively) at that point.

            Parameters
            ----------
            theta_phi: float or array of floats
                This argument may either contain an array with last dimension of
                size 2, where those two elements describe the (theta, phi) values in
                radians for each point; or it may contain just the theta values in
                radians, in which case the next argument must also be given.
            phi: None, float, or array of floats
                If this array is given, it must be able to broadcast against the
                first argument.

            Returns
            -------
            R: quaternion array
                If the second argument is not given to this function, the shape
                will be the same as the input shape except for the last dimension,
                which will be removed.  If the second argument is given, this
                output array will have the shape resulting from broadcasting the
                two input arrays against each other.

            """
            # Figure out the input angles from either type of input
            if phi is None:
                theta_phi = np.asarray(theta_phi)
                theta = theta_phi[..., 0]
                phi  = theta_phi[..., 1]
            else:
                theta = np.asarray(theta_phi)
                phi = np.asarray(phi)

            # Set up the output array
            R = np.empty(np.broadcast(theta, phi).shape + (4,), dtype=theta.dtype)

            # Compute the actual values of the quaternion components
            R[..., 0] =  np.cos(phi/2)*np.cos(theta/2)  # scalar quaternion components
            R[..., 1] = -np.sin(phi/2)*np.sin(theta/2)  # x quaternion components
            R[..., 2] =  np.cos(phi/2)*np.sin(theta/2)  # y quaternion components
            R[..., 3] =  np.sin(phi/2)*np.cos(theta/2)  # z quaternion components

            return cls(R)

    return mixin



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details:
# <https://github.com/moble/quaternionic/blob/master/LICENSE>

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:  # pragma: no cover
    import importlib_metadata

__version__ = importlib_metadata.version(__name__)

from .utilities import jit, guvectorize, float64, boolean
from . import algebra, properties, converters, distance, utilities

algebra_ufuncs = type('AlgebraUfuncs', (object,), dict())()
utilities.guvectorize_module_functions(algebra, algebra_ufuncs)

from .arrays import QuaternionicArray, array



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details:
# <https://github.com/moble/quaternionic/blob/master/LICENSE>

import numpy as np

from . import jit
from .properties import QuaternionPropertiesMixin
from .converters import QuaternionConvertersMixin


def QuaternionicArray(jit=jit, dtype=float):
    """Construct a quaternionic array type

    This factory returns a `class` encapsulating a quaternionic array type,
    where the jit function and dtype are passed to this factory function, and
    used when creating the class.  The returned class can then be used to
    instantiate actual arrays.  This allows us to, for example, skip jit
    compilation and construct numpy arrays of dtype `object`, so that we can
    use more general python types than the standard numeric types — such as
    sympy expressions.

    """
    class QArray(QuaternionPropertiesMixin(jit), QuaternionConvertersMixin(jit), np.ndarray):
        """Subclass of numpy arrays interpreted as quaternions

        This class encapsulates quaternion algebra, with numpy's "ufunc"s
        overridden by quaternionic methods.  Standard algebraic operations can be
        performed naturally — as in `q1+q2`, `q1*q2`, etc.  Numpy functions can
        also be used as expected — as in `np.exp(q)`, `np.log(q)`, etc.

        Because this is a subclass of numpy's ndarray object, its constructor takes
        anything the ndarray constructor takes, or just an ndarray to be considered
        as a quaternion array:

            q1 = quaternionic.array([1, 2, 3, 4])  # explicit array
            q2 = quaternionic.array(np.random.rand(10, 4))  # re-interpret ndarray

        In addition to the basic numpy array features, we also have a number of
        extra properties that are particularly useful for quaternions, including

          * Methods to extract and/or set components
            * w, x, y, z
            * i, j, k (equivalent to x, y, z)
            * scalar, vector (equivalent to w, [x, y, z])
            * real, imag (equivalent to scalar, vector)
          * Methods related to norms
            * abs (square-root of sum of squares of components)
            * norm (sum of squares of components)
            * modulus, magnitude (equal to abs)
            * absolute_square, abs2, mag2 (equal to norm)
            * normalized
            * conjugate, conj
            * inverse
          * Methods related to array infrastructure
            * ndarray
            * flattened
            * iterator

        There are also several converters to and from other representations of
        rotations, including

           * to/from_rotation_matrix
           * to_transformation_matrix
           * to/from_axis_angle
           * to/from_euler_angles
           * to/from_spherical_coordinates

        """

        # https://numpy.org/doc/1.18/user/basics.subclassing.html
        def __new__(cls, input_array, *args, **kwargs):
            from numbers import Real
            if isinstance(input_array, Real) and len(args) >= 3 and all(isinstance(a, Real) for a in args[:3]):
                input_array = [input_array,] + list(args[:3])
            input_array = np.asanyarray(input_array, dtype=dtype)
            if len(input_array.shape) == 0:
                raise ValueError("len(input_array.shape) == 0")
            if input_array.shape[-1] != 4:
                raise ValueError(
                    f"\nInput array has shape {input_array.shape} when viewed as a float array.\n"
                    "Its last dimension should have size 4, representing the components of a quaternion."
                )
            obj = input_array.view(cls)
            return obj

        def __getitem__(self, i):
            # Note that this simply assumes that if the returned array has last
            # dimension of size 4, it is a quaternionic array.  Obviously, this may
            # not always be true, but there's no simple way to decide more
            # correctly.
            r = super().__getitem__(i)
            return type(self)(r)

        def __array_finalize__(self, obj):
            if self.shape[-1] != 4:
                raise ValueError(
                    f"\nArray to finalize has shape {self.shape}; "
                    "last dimension should have size 4 to represent a quaternion.\n"
                    "If you are trying to slice the quaternions, you should append `.ndarray` before slicing.\n"
                    "For example, instead of `q[..., 2:]`, you must use `q.ndarray[..., 2:]` to return a\n"
                    "general (non-quaternion) numpy array.  This is a limitation of numpy.\n\n"
                    "Also note that quaternions have attributes like `q.w`, `q.x`, `q.y`, and `q.z` to return\n"
                    "arrays of individual components, and `q.vector` to return the \"vector\" part."
                )

        def __array_ufunc__(self, ufunc, method, *args, **kwargs):
            from . import algebra_ufuncs as algebra

            out = kwargs.pop('out', None)

            # We will not be supporting any more ufunc keywords beyond `out`
            if kwargs:
                raise NotImplementedError(f"Unrecognized arguments to {type(self).__name__}.__array_ufunc__: {kwargs}")

            if method in ["reduce", "accumulate", "reduceat", "outer", "at"]:
                raise NotImplementedError(f"Only __call__ method works for quaternionic arrays; got {method}")

            this_type = lambda o: isinstance(o, type(self))

            if ufunc in [
                np.add, np.subtract, np.multiply, np.divide, np.true_divide,
                np.bitwise_or, np.bitwise_xor, np.right_shift, np.left_shift,
            ]:
                # float64[4](float64[4], float64[4])
                if this_type(args[0]) and this_type(args[1]):
                    a1, a2 = args[:2]
                    b1 = a1.ndarray[..., 0]
                    b2 = a2.ndarray[..., 0]
                    shape = np.broadcast(b1, b2).shape + (4,)
                    result = out or np.zeros(shape, dtype=dtype)
                    if isinstance(result, tuple):
                        result = result[0]
                    result = result.view(np.ndarray)
                    getattr(algebra, ufunc.__name__)(a1.ndarray, a2.ndarray, result)
                    result = type(self)(result)

                # float64[4](float64, float64[4])
                elif not this_type(args[0]) and this_type(args[1]) and ufunc in [np.multiply, np.divide, np.true_divide]:
                    a1, a2 = args[:2]
                    b1 = a1
                    b2 = a2.ndarray[..., 0]
                    shape = np.broadcast(b1, b2).shape + (4,)
                    result = out or np.zeros(shape, dtype=dtype)
                    if isinstance(result, tuple):
                        result = result[0]
                    if isinstance(result, type(self)):
                        result = result.view(np.ndarray)
                    getattr(algebra, f"{ufunc.__name__}_scalar")(a1, a2.ndarray, result)
                    result = type(self)(result)

                # float64[4](float64[4], float64)
                elif this_type(args[0]) and not this_type(args[1]) and ufunc in [np.multiply, np.divide, np.true_divide]:
                    a1, a2 = args[:2]
                    b1 = a1.ndarray[..., 0]
                    b2 = a2
                    shape = np.broadcast(b1, b2).shape + (4,)
                    result = out or np.zeros(shape, dtype=dtype)
                    if isinstance(result, tuple):
                        result = result[0]
                    if isinstance(result, type(self)):
                        result = result.view(np.ndarray)
                    getattr(algebra, f"scalar_{ufunc.__name__}")(a1.ndarray, a2, result)
                    result = type(self)(result)
                else:
                    return NotImplemented

            # float64[4](float64[4])
            elif ufunc in [
                np.negative, np.positive, np.conj, np.conjugate, np.invert,
                np.exp, np.log, np.sqrt, np.square, np.reciprocal,
            ]:
                if this_type(args[0]):
                    a1 = args[0]
                    result = out or np.zeros(a1.shape, dtype=dtype)
                    if isinstance(result, tuple):
                        result = result[0]
                    if isinstance(result, type(self)):
                        result = result.view(np.ndarray)
                    getattr(algebra, ufunc.__name__)(a1.ndarray, result)
                    result = type(self)(result)
                else:
                    return NotImplemented

            # float64[4](float64[4], float64)
            elif ufunc in [np.float_power]:
                if this_type(args[0]) and not this_type(args[1]):
                    a1, a2 = args[:2]
                    b1 = a1.ndarray[..., 0]
                    b2 = a2
                    shape = np.broadcast(b1, b2).shape + (4,)
                    result = out or np.zeros(shape, dtype=dtype)
                    if isinstance(result, tuple):
                        result = result[0]
                    if isinstance(result, type(self)):
                        result = result.view(np.ndarray)
                    algebra.float_power(a1.ndarray, a2, result)
                    result = type(self)(result)
                else:
                    return NotImplemented

            # float64(float64[4])
            elif ufunc in [np.absolute]:
                if this_type(args[0]):
                    a1 = args[0]
                    result = out or np.zeros(a1.shape[:-1], dtype=dtype)
                    if isinstance(result, tuple):
                        result = result[0]
                    if isinstance(result, type(self)):
                        result = result.view(np.ndarray)
                    getattr(algebra, ufunc.__name__)(a1.ndarray, result)
                else:
                    return NotImplemented

            # bool(float64[4], float64[4])
            elif ufunc in [np.not_equal, np.equal, np.logical_and, np.logical_or]:
                # Note that these ufuncs are used in numerous unexpected places
                # throughout numpy, so we really need them for basic things to work
                if this_type(args[0]) and this_type(args[1]):
                    a1, a2 = args[:2]
                    b1 = a1.ndarray[..., 0]
                    b2 = a2.ndarray[..., 0]
                    shape = np.broadcast(b1, b2).shape
                    result = out or np.zeros(shape, dtype=bool)
                    if isinstance(result, tuple):
                        result = result[0]
                    if isinstance(result, type(self)):
                        result = result.view(np.ndarray)
                    getattr(algebra, ufunc.__name__)(a1.ndarray, a2.ndarray, result)
                else:
                    return NotImplemented

            # bool(float64[4])
            elif ufunc in [np.isfinite, np.isinf, np.isnan]:
                if this_type(args[0]):
                    a1 = args[0]
                    result = out or np.zeros(a1.shape[:-1], dtype=bool)
                    if isinstance(result, tuple):
                        result = result[0]
                    if isinstance(result, type(self)):
                        result = result.view(np.ndarray)
                    getattr(algebra, ufunc.__name__)(a1.ndarray, result)
                else:
                    return NotImplemented

            else:
                return NotImplemented

            return result

        def __repr__(self):
            return 'quaternionic.' + repr(self.ndarray)

        def __str__(self):
            return 'quaternionic.' + str(self.ndarray)

    return QArray


array = QuaternionicArray()



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details:
# <https://github.com/moble/quaternionic/blob/master/LICENSE>

import abc
import numpy as np
from . import jit


def integrate_angular_velocity(ω, t, axis=0):
    raise NotImplementedError()

def slerp():
    raise NotImplementedError()

def squad():
    raise NotImplementedError()


class QuaternionTimeseriesMixin(abc.ABC):
    """Timeseries methods for quaternionic array class"""

    def spline(self, t, axis=0):
        """Construct a spline object for the quaternion data

        Parameters
        ==========
        t: array-like of float
            This array represents the times at which the quaternions are
            measured, and with respect to which the derivative will be taken.
            Note that these times must be real, finite and in strictly
            increasing order.
        axis: int, optional
            Axis along which this array is assumed to be varying with
            `t`. Meaning that for t[i] the corresponding quaternions are
            `np.take(self, i, axis=axis)`.  Defaults to 0.

        """
        from scipy.interpolate import CubicSpline
        spline = CubicSpline(t, self.ndarray, axis=axis)
        return spline

    def derivative(self, t, t_new=None, axis=0, nu=1):
        """Take derivative of quaternion timeseries with respect to `t`

        Parameters
        ==========
        t: array-like of float
            This array represents the times at which the quaternions are
            measured, and with respect to which the derivative will be taken.
            Note that these times must be real, finite and in strictly
            increasing order.
        t_new: array-like of float, optional
            If present, the output is interpolated to this set of times.
            Defaults to None, meaning that the original set of times will be
            used.
        axis: int, optional
            Axis along which this array is assumed to be varying with
            `t`. Meaning that for t[i] the corresponding quaternions are
            `np.take(self, i, axis=axis)`.  Defaults to 0.
        nu: int, optional
            Order of derivative to evaluate. Default is 1, i.e., compute the
            first derivative. If negative, the antiderivative is returned.

        """
        t_new = t_new or t
        spline = self.spline(t, axis=axis)
        return type(self)(spline.derivative(nu=nu)(t_new))
        
    def angular_velocity(self, t, t_new=None, axis=0):
        """Compute the angular velocity of quaternion timeseries with respect to `t`

        Note that, for both unit and non-unit quaternions `Q`, we define the
        angular velocity as

            ω = 2 * dQ/dt * Q⁻¹

        This agress with the standard definition when `Q` is a unit quaternion
        and we rotate a vector `v` according to

            v' = Q * v * Q̄ = Q * v * Q⁻¹,

        in which case ω is a "pure vector" quaternion, and we have the usual

            dv'/dt = ω × v'.

        It also generalizes this to the case where `Q` is not a unit
        quaternion, which means that it also rescales the vector by the amount
        Q*Q̄.  In this case, ω also has a scalar component encoding twice the
        logarithmic time-derivative of this rescaling, and we have

            dv'/dt = ω * v' + v' * ω̄.

        Parameters
        ==========
        t: array-like of float
            This array represents the times at which the quaternions are
            measured, and with respect to which the derivative will be taken.
            Note that these times must be real, finite and in strictly
            increasing order.
        t_new: array-like of float, optional
            If present, the output is interpolated to this set of times.
            Defaults to None, meaning that the original set of times will be
            used.
        axis: int, optional
            Axis along which this array is assumed to be varying with
            `t`. Meaning that for t[i] the corresponding quaternions are
            `np.take(self, i, axis=axis)`.  Defaults to 0.

        """
        spline = self.spline(t, axis=axis)
        if t_new is None:
            Q = self
        else:
            Q = type(self)(spline(t_new))
        t_new = t_new or t
        Q̇ = type(self)(spline.derivative()(t_new))
        return 2 * Q̇ / Q

    def minimal_rotation(self, t, t_new=None, axis=0, iterations=2):
        """Adjust frame so that there is no rotation about z' axis

        The output of this function is a frame that rotates the z axis onto the
        same z' axis as the input frame, but with minimal rotation about that
        axis.  This is done by pre-composing the input rotation with a rotation
        about the z axis through an angle γ, where

            dγ/dt = 2*(dR/dt * z * R̄).w

        This ensures that the angular velocity has no component along the z'
        axis.

        Note that this condition becomes easier to impose the closer the input
        rotation is to a minimally rotating frame, which means that repeated
        application of this function improves its accuracy.  By default, this
        function is iterated twice, though a few more iterations may be called
        for.

        Parameters
        ==========
        t: array-like of float
            This array represents the times at which the quaternions are
            measured, and with respect to which the derivative will be taken.
            Note that these times must be real, finite and in strictly
            increasing order.
        t_new: array-like of float, optional
            If present, the output is interpolated to this set of times.
            Defaults to None, meaning that the original set of times will be
            used.
        axis: int, optional
            Axis along which this array is assumed to be varying with
            `t`. Meaning that for t[i] the corresponding quaternions are
            `np.take(self, i, axis=axis)`.  Defaults to 0.
        iterations: int, optional
            Repeat the minimization to refine the result.  Defaults to 2.

        """
        from scipy.interpolate import CubicSpline
        if iterations == 0:
            return self
        z = type(self)([0.0, 0.0, 0.0, 1.0])
        t_new = t_new or t
        spline = self.spline(t, axis=axis)
        R = type(self)(spline(t_new))
        Rdot = type(self)(spline.derivative()(t_new))
        γ̇over2 = (Rdot * z * np.conjugate(R)).w
        γover2 = CubicSpline(t_new, γ̇over2).antiderivative()(t_new)
        Rγ = np.exp(z * γover2)
        return (R * Rγ).minimal_rotation(t_new, t_new=None, axis=axis, iterations=iterations-1)

    def slerp(self):
        """Spherical linear interpolation of quaternions

        See also
        --------
        squad: spherical quadratic interpolation
        quaternionic.slerp_pairwise

        """
        raise NotImplementedError()

    def squad(self):
        """Spherical quadratic interpolation of quaternions

        See also
        --------
        slerp: spherical linear interpolation
        quaternionic.squad_pairwise

        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['quaternionic']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.13,<2.0', 'scipy>=1.0,<2.0']

extras_require = \
{':implementation_name == "cpython"': ['numba>=0.50'],
 ':python_version < "3.8"': ['importlib-metadata>=1.0,<2.0'],
 'mkdocs:implementation_name == "cpython"': ['mkdocs>=1.1.2'],
 'mktheapidocs:implementation_name == "cpython"': ['mktheapidocs>=0.2'],
 'pymdown-extensions:implementation_name == "cpython"': ['pymdown-extensions>=8,<9']}

setup_kwargs = {
    'name': 'quaternionic',
    'version': '0.1.10',
    'description': 'Interpret numpy arrays as quaternionic arrays with numba acceleration',
    'long_description': '[![Test Status](https://github.com/moble/quaternionic/workflows/tests/badge.svg)](https://github.com/moble/quaternionic/actions)\n[![Test Coverage](https://codecov.io/gh/moble/quaternionic/branch/master/graph/badge.svg)](https://codecov.io/gh/moble/quaternionic)\n[![Documentation Status](https://readthedocs.org/projects/quaternionic/badge/?version=latest)](https://quaternionic.readthedocs.io/en/latest/?badge=latest)\n[![PyPI Version](https://img.shields.io/pypi/v/quaternionic?color=)](https://pypi.org/project/quaternionic/)\n[![Conda Version](https://img.shields.io/conda/vn/conda-forge/quaternionic.svg?color=)](https://anaconda.org/conda-forge/quaternionic)\n\n\n# Quaternionic arrays\n\nThis module subclasses numpy\'s array type, interpreting the array as an array of quaternions, and\naccelerating the algebra using numba.  This enables natural manipulations, like multiplying\nquaternions as `a*b`, while also working with standard numpy functions, as in `np.log(q)`.  There is\nalso basic initial support for symbolic manipulation of quaternions by creating quaternionic arrays\nwith sympy symbols as elements, though this is a work in progress.\n\nThis package has evolved from the [quaternion](https://github.com/moble/quaternion) package, which\nadds a quaternion dtype directly to numpy.  In some ways, that is a better approach because dtypes\nare built in to numpy, making it more robust than this package.  However, that approach has its own\nlimitations, including that it is harder to maintain, and requires much of the code to be written in\nC, which also makes it harder to distribute.  This package is written entirely in python code, but\nshould actually have comparable performance because it is compiled by numba.  Moreover, because the\ncore code is written in pure python, it is reusable for purposes other than the core purpose of this\npackage, which is to provide the numeric array type.\n\n\n# Installation\n\nBecause this package is pure python code, installation is very simple.  In particular, with\na reasonably modern installation, you can just run\n\n```bash\nconda install -c conda-forge quaternionic\n```\n\nor\n\n```bash\npip install quaternionic\n```\n\nThese will download and install the package.  It can also be installed from source by running `pip\ninstall .` from the top-level directory of the code.\n\nNote that only python 3.6 or greater is supported, and pip itself must be version 10.0 or greater.\n(I have tried to support PyPy3, although I cannot test this as `scipy` does not currently install.\nPull requests are welcome.)  In any case, I strongly recommend installing by way of an environment\nmanager — especially [conda](https://docs.anaconda.com/anaconda/install/), though other managers\nlike `virtualenv` or `pipenv` should also work.\n\nFor development work, the best current option is [poetry](https://python-poetry.org/).  From the\ntop-level directory, run `poetry install` or just `poetry run <some command>`.\n\n\n# Usage\n\n## Basic construction\n\nThe key function is `quaternionic.array`, which takes nearly the same arguments as `numpy.array`,\nexcept that whatever array will result must have a final axis of size 4 (and the `dtype` must be\n`float`).  As long as these conditions are satisfied, we can create new arrays or just reinterpret\nexisting arrays:\n\n```python\nimport numpy as np\nimport quaternionic\n\na = 1.0 - np.random.rand(17, 11, 4)  # Just some random numbers; last dimension is 4\nq1 = quaternionic.array(a)  # Reinterpret an existing array\nq2 = quaternionic.array([1.2, 2.3, 3.4, 4.5])  # Create a new array\n```\n\nIn this example, `q1` is an array of 187 (17*11) quaternions, just to demonstrate that any number of\ndimensions may be used, as long as the final dimension has size 4.\n\nHere, the original array `a` will still exist just as it was, and will behave just as a normal numpy\narray — including changing its values (which will change the values in `q1`), slicing, math, etc.\nHowever, `q1` will be another\n["view"](https://numpy.org/doc/stable/reference/generated/numpy.ndarray.view.html) into the same\ndata.  Operations on `q1` will be quaternionic.  For example, whereas `1/a` returns the element-wise\ninverse of each float in the array, `1/q1` returns the *quaternionic* inverse of each quaternion.\nSimilarly, if you multiply two quaternionic arrays, their product will be computed with the usual\nquaternion multiplication, rather than element-wise multiplication of floats as numpy usually\nperforms.\n\n## Algebra\n\nAll the usual quaternion operations are available, including\n\n  * Addition `q1 + q2`\n  * Subtraction `q1 - q2`\n  * Multiplication `q1 * q2`\n  * Division `q1 / q2`\n  * Scalar multiplication `q1 * s == s * q1`\n  * Scalar division `q1 / s` and `s / q1`\n  * Reciprocal `np.reciprocal(q1) == 1/q1`\n  * Exponential `np.exp(q1)`\n  * Logarithm `np.log(q1)`\n  * Square-root `np.sqrt(q1)`\n  * Conjugate `np.conjugate(q1) == np.conj(q1)`\n\nAll numpy [ufuncs](https://numpy.org/doc/stable/reference/ufuncs.html) that make sense for\nquaternions are supported.  When the arrays have different shapes, the usual numpy\n[broadcasting](https://numpy.org/doc/stable/user/basics.broadcasting.html) rules take effect.\n\n\n## Attributes\n\nIn addition to the basic numpy array features, we also have a number of extra properties that are\nparticularly useful for quaternions, including\n\n  * Methods to extract and/or set components\n    * `w`, `x`, `y`, `z`\n    * `i`, `j`, `k` (equivalent to `x`, `y`, `z`)\n    * `scalar`, `vector` (equivalent to `w`, [`x`, `y`, `z`])\n    * `real`, `imag` (equivalent to `scalar`, `vector`)\n  * Methods related to norms\n    * `abs` (square-root of sum of squares of components)\n    * `norm` (sum of squares of components)\n    * `modulus`, `magnitude` (equal to `abs`)\n    * `absolute_square`, `abs2`, `mag2` (equal to `norm`)\n    * `normalized`\n    * `inverse`\n  * Methods related to array infrastructure\n    * `ndarray` (the numpy array underlying the quaternionic array)\n    * `flattened` (all dimensions but last are flattened into one)\n    * `iterator` (iterate over all quaternions)\n\nNote that this package makes a distinction between `abs` and `norm` — the latter being equal to the\nsquare of the former.  This version of the norm is also known as the "Cayley" norm, commonly used\nwhen emphasizing the properties of an object in an algebra, as opposed to the "Euclidean" norm more\ncommon when emphasizing the properties of an object in a vector space — though of course, algebras\nare vector spaces with additional structure.  This choice agrees with the [Boost library\'s\nimplementation of\nquaternions](https://www.boost.org/doc/libs/1_74_0/libs/math/doc/html/math_toolkit/value_op.html),\nas well as this package\'s forerunner\n[quaternion](https://github.com/moble/quaternion/blob/99913120b1b2a8a5eb7769c29ee760a236d40880/quaternion.h#L115-L120).\nThis also agrees with the corresponding functions on the [C++ standard library\'s complex\nnumbers](http://www.cplusplus.com/reference/complex/norm/).  Because this may be confusing, a number\nof aliases are also provided that may be less confusing.  For example, some people find the pair\n`abs` and `abs2` (meaning the square of `abs`) to be more sensible.\n\n\n## Rotations\n\nThe most common application of quaternions is to representing rotations by means of unit\nquaternions.  Note that this package does not *restrict* quaternions to have unit norms, since it is\nusually better for numerical purposes not to do so.  For example, whereas rotation of a vector $v$\nby a quaternion is usually implemented as $R\\, v\\, \\bar{R}$, it is generally better to drop the\nassumption that the quaternion has unit magnitude and implement rotation as $R\\, v\\, R^{-1}$.  This\nis almost always more efficient, and more accurate.  That is what this package does by default\nwhenever rotations are involved.\n\nAlthough this package does not restrict to unit quaternions, there are several converters to and\nfrom other representations of rotations, including\n\n   * `to_rotation_matrix`, `from_rotation_matrix`\n   * `to_transformation_matrix` (for non-unit quaternions)\n   * `to_axis_angle`, `from_axis_angle` representation\n   * `to_euler_angles`, `from_euler_angles` (though using Euler angles is almost always a bad idea)\n   * `to_spherical_coordinates`, `from_spherical_coordinates`\n   * `to_angular_velocity`, `from_angular_velocity`\n\nNote that the last item relates to quaternion-valued functions of time.  Converting to an angular\nvelocity requires differentiation, while converting from angular velocity requires integration (as\nexplored in [this paper](https://arxiv.org/abs/1604.08139)).\n\nFor these converters, the "to" functions are properties on the individual arrays, whereas the "from"\nfunctions are "classmethod"s that take the corresponding objects as inputs.  For example, we could\nwrite\n\n```python\nq1 = quaternionic.array(np.random.rand(100, 4)).normalized\nm = q1.to_rotation_matrix\n```\n\nto obtain the matrix `m` *from* a quaternionic array `q1`.  (Here, `m` is actually a series of 100\n3x3 matrices corresponding to the 100 quaternions in `q1`.)  On the other hand, to obtain a\nquaternionic array from some matrix `m`, we would write\n\n```python\nq2 = quaternionic.array.from_rotation_matrix(m)\n```\n\nAlso note that, because the unit quaternions form a "double cover" of the rotation group (meaning\nthat quaternions `q` and `-q` represent the same rotation), these functions are not perfect inverses\nof each other.  In this case, for example, `q1` and `q2` may have opposite signs.  We can, however,\nprove that these quaternions represent the same rotations by measuring the "distance" between the\nquaternions as rotations:\n\n```python\nnp.max(quaternionic.distance.rotation.intrinsic(q1, q2))  # Typically around 1e-15\n```\n\n\n## Distance functions\n\nThe `quaternionic.distance` contains four distance functions:\n\n  * `rotor.intrinsic`\n  * `rotor.chordal`\n  * `rotation.intrinsic`\n  * `rotation.chordal`\n\nThe "rotor" distances do not account for possible differences in signs, meaning that rotor distances\ncan be large even when they represent identical rotations; the "rotation" functions just return the\nsmaller of the distance between `q1` and `q2` or the distance between `q1` and `-q2`.  So, for\nexample, either "rotation" distance between `q` and `-q` is always zero, whereas neither "rotor"\ndistance between `q` and `-q` will ever be zero (unless `q` is zero).  The "intrinsic" functions\nmeasure the geodesic distance within the manifold of *unit* quaternions, and is somewhat slower but\nmay be more meaningful; the "chordal" functions measure the Euclidean distance in the (linear) space\nof all quaternions, and is faster but its precise value is not necessarily as meaningful.\n\nThese functions satisfy some important conditions.  For each of these functions `d`, and for any\nnonzero quaternions `q1` and `q2`, and *unit* quaternions `q3` and `q4`, we have\n\n  * symmetry: `d(q1, q2) = d(q2, q1)`\n  * invariance: `d(q3*q1, q3*q2) = d(q1, q2) = d(q1*q4, q2*q4)`\n  * identity: `d(q1, q1) = 0`\n  * positive-definiteness:\n    * For rotor functions `d(q1, q2) > 0` whenever `q1 ≠ q2`\n    * For rotation functions `d(q1, q2) > 0` whenever `q1 ≠ q2` and `q1 ≠ -q2`\n\nNote that the rotation functions also satisfy both the usual identity property `d(q1, q1) = 0` and\nthe opposite-identity property `d(q1, -q1) = 0`.\n\nSee [Moakher (2002)](https://doi.org/10.1137/S0895479801383877) for a nice general discussion.\n\n\n## Interpolation\n\nFinally, there are also capabilities related to interpolation\n\n  * slerp (spherical linear interpolation)\n  * squad (spherical quadratic interpolation)\n\n\n# Related packages\n\nOther python packages with some quaternion features include\n\n  * [quaternion](https://github.com/moble/quaternion/) (core written in C; very fast; adds\n    quaternion `dtype` to numpy; named\n    [numpy-quaternion](https://pypi.org/project/numpy-quaternion/) on pypi due to name conflict)\n  * [clifford](https://github.com/pygae/clifford) (very powerful; more general geometric algebras)\n  * [rowan](https://github.com/glotzerlab/rowan) (many features; similar approach to this package;\n    no acceleration or overloading)\n  * [pyquaternion](http://kieranwynn.github.io/pyquaternion/) (many features; pure python; no\n    acceleration or overloading)\n  * [quaternions](https://github.com/mjsobrep/quaternions) (basic pure python package; no\n    acceleration; specialized for rotations only)\n  * [scipy.spatial.transform.Rotation.as\\_quat](https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.transform.Rotation.as_quat.html)\n    (quaternion output for `Rotation` object)\n  * [mathutils](https://gitlab.com/ideasman42/blender-mathutils) (a Blender package with python\n    bindings)\n  * [Quaternion](https://pypi.org/project/Quaternion/) (extremely limited capabilities; unmaintained)\n\nAlso note that there is some capability to do symbolic manipulations of quaternions in these\npackages:\n\n  * [galgebra](https://github.com/pygae/galgebra) (more general geometric algebras; analogous to\n    `clifford`, but for symbolic calculations)\n  * [sympy.algebras.quaternion](https://docs.sympy.org/latest/modules/algebras.html)\n',
    'author': 'Michael Boyle',
    'author_email': 'michael.oliver.boyle@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/moble/quaternionic',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
